﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Windows.Forms;

namespace SCDP
{
    // (c) 2018 radiolistener
    // http://www.eevblog.com/forum/testgear/siglent-sds1000x-how-to-make-direct-ethernet-connection/
    public partial class FormMain : Form
    {
        private const string TITLE = "SCDP";
        private Oscilloscope _device;
        
        public FormMain()
        {
            InitializeComponent();
            UpdateState();
        }

        private void btnConnect_Click(object sender, EventArgs e)
        {
            btnConnect.Enabled = false;
            try
            {
                if (_device == null)
                {
                    var session = new SessionVxi11(IPAddress.Parse(txtIP.Text));
                    session.TimeoutMilliseconds = 1000 * 60;
                    session.Clear();
                    _device = new Oscilloscope(session);
                }
                else
                {
                    _device.Dispose();
                    _device = null;
                }
                if (_device != null)
                {
                    Text = string.Format("[{0}] {1}", TITLE, _device.GetIDN());                    
                    var screen = _device.GetScreenDump();
                    using (var stream = new MemoryStream(screen))
                    {
                        pbxScreen.Image = new Bitmap(stream);
                    }
                }
            }
            finally
            {
                btnConnect.Enabled = true;
                UpdateState();
            }
        }

        private void UpdateState()
        {
            if (_device == null)
            {
                txtIP.Enabled = true;
                btnConnect.Text = "Connect";
                btnRefresh.Enabled = false;
                Text = string.Format("[{0}]", TITLE);
            }
            else
            {
                txtIP.Enabled = false;
                btnConnect.Text = "Disconnect";
                btnRefresh.Enabled = true;
                Text = string.Format("[{0}] {1}", TITLE, _device.GetIDN());
            }
        }

        private void btnRefresh_Click(object sender, EventArgs e)
        {
            btnRefresh.Enabled = false;
            try
            {
                DimScreen();
                if (_device == null) return;
                var screen = _device.GetScreenDump();
                using (var stream = new MemoryStream(screen))
                {
                    pbxScreen.Image = new Bitmap(stream);
                }
            }
            finally
            {
                btnRefresh.Enabled = true;
            }
        }

        private void DimScreen()
        {
            var image = pbxScreen.Image;
            var bmp = new Bitmap(image.Width, image.Height, image.PixelFormat);
            using (var g = Graphics.FromImage(bmp))
            {
                var cm = new ColorMatrix(new float[][] { 
                    new float[] {1, 0, 0, 0, 0},
                    new float[] {0, 1, 0, 0, 0},
                    new float[] {0, 0, 1, 0, 0},
                    new float[] {.2f, .2f, .2f, 1, 0}, 
                    new float[] {0, 0, 0, 0, 1}});
                var imgAttrs = new ImageAttributes();
                imgAttrs.SetColorMatrix(cm);
                var rect = new Rectangle(0, 0, image.Width, image.Height);
                g.DrawImage(image, rect, 0, 0, image.Width, image.Height, 
                    GraphicsUnit.Pixel, imgAttrs);
            }
            pbxScreen.Image = bmp;
            pbxScreen.Refresh();
        }

        private SaveFileDialog _saveDialog;

        private void btnSave_Click(object sender, EventArgs e)
        {
            if (pbxScreen.Image == null) return;
            if (_saveDialog == null)
            {
                _saveDialog = new SaveFileDialog();
                _saveDialog.InitialDirectory = ".";
                _saveDialog.AddExtension = true;
                _saveDialog.DefaultExt = ".png";
                _saveDialog.Filter = "PNG file (*.png)|*.png|BMP file (*.bmp)|*.bmp";
                _saveDialog.FilterIndex = 0;
            }
            _saveDialog.FileName = null;
            if (_saveDialog.ShowDialog() != DialogResult.OK)
                return;
            var ext = Path.GetExtension(_saveDialog.FileName);
            var format = ImageFormat.Bmp;
            switch (ext.ToLowerInvariant())
            {
                case ".png": format = ImageFormat.Png; break;
                case ".bmp": format = ImageFormat.Bmp; break;
                case ".jpeg":
                case ".jpg": format = ImageFormat.Jpeg; break;
                case ".gif": format = ImageFormat.Gif; break;
                case ".tif":
                case ".tiff": format = ImageFormat.Tiff; break;
            }
            using (var bmp = new Bitmap(pbxScreen.Image))
            {
                bmp.Save(_saveDialog.FileName, format);
            }
        }
    }
}
