﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;


namespace SCDP
{
    // (c) 2018 radiolistener
    // http://www.eevblog.com/forum/testgear/siglent-sds1000x-how-to-make-direct-ethernet-connection/
    public class Oscilloscope : IDisposable
    {
        private readonly ISessionVxi11 _session;

        #region .ctor

        public Oscilloscope(ISessionVxi11 session)
        {
            _session = session;
        }

        public void Dispose()
        {
            _session.Dispose();
        }

        #endregion .ctor



        #region Read/Write

        //[DebuggerStepThrough]
        public void WriteCommand(string cmd)
        {
            _session.RawWrite(cmd);
        }

        //[DebuggerStepThrough]
        public string ReadString()
        {
            return _session.RawReadString();
        }

        public byte[] ReadData(int length)
        {
            ReadStatus status;
            var readBytes = _session.RawRead((uint)length, out status);
            Trace.WriteLine(string.Format("Readed {0} bytes from device with Status {1}", readBytes.Length, status));
            return readBytes;
        }

        #endregion Read/Write


        #region Commands

        //[DebuggerStepThrough]
        public string GetIDN()
        {
            WriteCommand("*IDN?");
            return ReadString();
            //session.RawIO.Write("C1:ATTN 1");         // probe attn = 1, check "C1:ATTN?"
            //session.RawIO.Write("C1:ACQW SAMPLING");  // acq mode: normal
        }

        public byte[] GetScreenDump()
        {
            WriteCommand("SCDP");       // SCREEN_DUMP
            return ReadData(64000000);//1152057);
        }

        //[DebuggerStepThrough]
        public byte[] GetWaveform(int channel)
        {
            if (channel < 1 || channel > 4) return null;
            //C1:WF? DAT2
            //C1:WF ALL,#9014000000
            WriteCommand(string.Format("C{0}:WF? DAT2", channel));
            var data = ReadData(128000000);
            var pos = -1;
            for (var i=0; i < data.Length && i < 64; i++)
            {
                if (data[i] == '#')
                {
                    pos = i+1;
                    break;
                }
            }
            if (pos < 0 || 
                pos >= data.Length || 
                !char.IsDigit((char)data[pos])) 
            {
                return null;
            }
            var len = data[pos] - '0';
            pos++;
            if (pos + len > data.Length) return null;
            var lenStr = Encoding.ASCII.GetString(data, pos, len);
            pos += len;
            if (!lenStr.All(char.IsDigit)) return null;
            len = int.Parse(lenStr);
            var subData = new byte[len];
            Array.Copy(data, pos, subData, 0, len);
            return subData;
        }

        #endregion Commands
    }
}
