﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;

namespace SCDP
{
    // (c) 2018 radiolistener
    // http://www.eevblog.com/forum/testgear/siglent-sds1000x-how-to-make-direct-ethernet-connection/
    public class SessionVxi11 : ISessionVxi11
    {
        private readonly uint _linkId;
        private TcpClient _tcpClient;
        private uint _xid;

        private int _timeoutMilliseconds;
        public int TimeoutMilliseconds 
        {
            get { return _timeoutMilliseconds; }
            set 
            {
                _timeoutMilliseconds = value;
                if (_tcpClient != null)
                {
                    _tcpClient.Client.ReceiveTimeout = value;
                    _tcpClient.Client.SendTimeout = value;
                }
            }
        }

        public SessionVxi11(IPAddress ip)
        {
            _xid = 0;
            TimeoutMilliseconds = 2000;
            
            _tcpClient = new TcpClient();
            _tcpClient.Client.Blocking = true;
            _tcpClient.Client.ReceiveTimeout = TimeoutMilliseconds;
            _tcpClient.Client.SendTimeout = TimeoutMilliseconds;
            _tcpClient.Connect(new IPEndPoint(ip, 111));

            var rpcPortmap = CreateRpcGetPort(MakeXid());
            _tcpClient.Client.Send(rpcPortmap);

            // RPC
            var rpcLength = RxUInt32();
            var rpcXid = RxUInt32();
            var msgType = RxUInt32();       // Message Type: Reply (1)
            var replyState = RxUInt32();    // Reply state: accepted (0)
            var verifFlavor = RxUInt32();   // Verifier Flavor: AUTH_NULL (0)
            var verifLength = RxUInt32();   // Verifier Length: 0
            var acceptState = RxUInt32();   // Accept State: RPC executed successfuly (0)
            // Portmap
            var port = RxUInt32();          // Port: 9009

            _tcpClient.Close();
            _tcpClient = new TcpClient();
            _tcpClient.Client.Blocking = true;
            _tcpClient.Client.ReceiveTimeout = TimeoutMilliseconds;
            _tcpClient.Client.SendTimeout = TimeoutMilliseconds;
            _tcpClient.Connect(new IPEndPoint(ip, (int)port));

            var rpc = CreateRpcLinkCreate(MakeXid(), 0x00000000, "inst0");
            _tcpClient.Client.Send(rpc);

            // RPC
            rpcLength = RxUInt32();
            rpcXid = RxUInt32();
            msgType = RxUInt32();       // Message Type: Reply (1)
            replyState = RxUInt32();    // Reply state: accepted (0)
            verifFlavor = RxUInt32();   // Verifier Flavor: AUTH_NULL (0)
            verifLength = RxUInt32();   // Verifier Length: 0
            acceptState = RxUInt32();   // Accept State: RPC executed successfuly (0)
            // VXI-11
            var errorCode = RxUInt32();     // Error Code: No Error (0)
            _linkId = RxUInt32();           // Link ID: 11
            var abortPort = RxUInt32();     // Abort Port: 1
            var maxRxSize = RxUInt32();     // Maximum Receive Size: 4096
        }

        public void Dispose()
        {
            var rpc = CreateRpcLinkDestroy(MakeXid(), _linkId);
            _tcpClient.Client.Send(rpc);

            // RPC
            var rpcLength = RxUInt32();
            var rpcXid = RxUInt32();
            var msgType = RxUInt32();       // Message Type: Reply (1)
            var replyState = RxUInt32();    // Reply state: accepted (0)
            var verifFlavor = RxUInt32();   // Verifier Flavor: AUTH_NULL (0)
            var verifLength = RxUInt32();   // Verifier Length: 0
            var acceptState = RxUInt32();   // Accept State: RPC executed successfuly (0)
            // VXI-11
            var errorCode = RxUInt32();     // Error Code: No Error (0)
        }

        public void Clear()
        {
            var rpc = CreateRpcClear(MakeXid(), _linkId, (uint)TimeoutMilliseconds);
            _tcpClient.Client.Send(rpc);

            // RPC
            var rpcLength = RxUInt32();
            var rpcXid = RxUInt32();
            var msgType = RxUInt32();       // Message Type: Reply (1)
            var replyState = RxUInt32();    // Reply state: accepted (0)
            var verifFlavor = RxUInt32();   // Verifier Flavor: AUTH_NULL (0)
            var verifLength = RxUInt32();   // Verifier Length: 0
            var acceptState = RxUInt32();   // Accept State: RPC executed successfuly (0)
            // VXI-11
            var errorCode = RxUInt32();     // Error Code: No Error (0)
        }

        public void RawWrite(string buffer)
        {
            var data = Encoding.ASCII.GetBytes(buffer);
            RawWrite(data, 0, data.Length);
        }

        public string RawReadString()
        {
            ReadStatus readStatus;
            var data = RawRead(1024, out readStatus);
            return Encoding.ASCII.GetString(data);
        }

        public void RawWrite(byte[] buffer, int index, int length)
        {
            var rpc = CreateRpcWrite(MakeXid(), _linkId, buffer, index, length, (uint)TimeoutMilliseconds);
            _tcpClient.Client.Send(rpc);

            // RPC
            var rpcLength = RxUInt32();
            var rpcXid = RxUInt32();
            var msgType = RxUInt32();       // Message Type: Reply (1)
            var replyState = RxUInt32();    // Reply state: accepted (0)
            var verifFlavor = RxUInt32();   // Verifier Flavor: AUTH_NULL (0)
            var verifLength = RxUInt32();   // Verifier Length: 0
            var acceptState = RxUInt32();   // Accept State: RPC executed successfuly (0)
            // VXI-11
            var errorCode = RxUInt32();     // Error Code: No Error (0)
            var vxiSize = RxUInt32();       // Size: 5
        }

        public byte[] RawRead(uint count, out ReadStatus readStatus)
        {
            readStatus = ReadStatus.Unknown;
            byte[] array = new byte[count];
            uint actualRead;
            RawReadInt(array, 0L, count, out actualRead, out readStatus);
            if (actualRead < count)
            {
                Array.Resize<byte>(ref array, (int)actualRead);
            }
            return array;            
        }

        private void RawReadInt(byte[] buffer, long index, uint count, out uint actualCount, out ReadStatus readStatus)
        {
            readStatus = ReadStatus.Unknown;
            actualCount = 0;

            var rpc = CreateRpcRead(MakeXid(), _linkId, count, (uint)TimeoutMilliseconds);
            _tcpClient.Client.Send(rpc);

            // RPC
            var rpcLength = RxUInt32();
            var rpcXid = RxUInt32();
            var msgType = RxUInt32();       // Message Type: Reply (1)
            var replyState = RxUInt32();    // Reply state: accepted (0)
            var verifFlavor = RxUInt32();   // Verifier Flavor: AUTH_NULL (0)
            var verifLength = RxUInt32();   // Verifier Length: 0
            var acceptState = RxUInt32();   // Accept State: RPC executed successfuly (0)
            // VXI-11
            var errorCode = RxUInt32();    // Error Code: No Error (0)
            var reason = RxUInt32();       // Reason: 0x00000004 (END)
            var dataLength = RxUInt32();   // Data length
            
            var partLength = rpcLength & 0x7fffffff;
            partLength -= 36;
            var tmp = new byte[0x10000];
            do
            {
                if (tmp.Length < partLength)
                    tmp = new byte[partLength];
                //if (actualCount+partLength > count)
                //{
                //    partLength = count - actualCount;
                //    readStatus = ReadStatus.MaximumCountReached;
                //}
                _tcpClient.Client.Receive(tmp, (int)partLength, SocketFlags.None);
                Array.Copy(tmp, 0, buffer, index + actualCount, partLength);
                actualCount += partLength;
                if ((rpcLength & 0x80000000) != 0)
                {
                    readStatus = ReadStatus.EndReceived;
                    break;
                }
                rpcLength = RxUInt32();
                partLength = rpcLength & 0x7fffffff;
            } while (actualCount < dataLength);
        }

        private uint MakeXid()
        {
            _xid++;
            return _xid;
        }

        private uint RxUInt32()
        {
            var buffer = new byte[4];
            _tcpClient.Client.Receive(buffer);
            return ReadUInt32(buffer, 0);
        }

        private static byte[] CreateRpcGetPort(uint xid)
        {
            var list = new List<byte>();
            // Portmap VXI-11
            WriteUInt32(list, 395183);      // Program: VXI-11 Core (395183)
            WriteUInt32(list, 1);           // Version: 1
            WriteUInt32(list, 6);           // Proto: TCP (6)
            WriteUInt32(list, 0);           // Port: 0
            var vxi = list.ToArray();
            list.Clear();
            // RPC
            uint rpcLength = (uint)(40 + vxi.Length);
            rpcLength |= 0x80000000;              // last fragment: yes

            WriteUInt32(list, rpcLength);
            WriteUInt32(list, xid);
            WriteUInt32(list, 0);       // Message Type: Call (0)
            WriteUInt32(list, 2);       // RPC Version: 2
            WriteUInt32(list, 100000);  // Program: Portmap (100000)
            WriteUInt32(list, 2);       // Program Version: 2
            WriteUInt32(list, 3);       // Procedure: GETPORT (3)
            WriteUInt32(list, 0);       // Credentials flavor: 0
            WriteUInt32(list, 0);       // Credentials length: 0
            WriteUInt32(list, 0);       // Verifier flavor: 0
            WriteUInt32(list, 0);       // Verifier length: 0
            list.AddRange(vxi);
            return list.ToArray();
        }

        // clientId=0xc0a80121
        // deviceName="inst0"
        private static byte[] CreateRpcLinkCreate(uint xid, uint clientId, string deviceName)
        {
            var list = new List<byte>();
            // VXI-11
            WriteUInt32(list, clientId);      // Client ID: 3232235809 (0xC0A80121)
            WriteUInt32(list, 0);             // Lock Device: No
            WriteUInt32(list, 0);             // Lock Timeout: 0
            var nameData = Encoding.ASCII.GetBytes(deviceName);
            WriteUInt32(list, (uint)nameData.Length);// Device Name length
            list.AddRange(nameData);
            for (var i = 0; i < 4 - (nameData.Length % 4); i++)
            {
                list.Add(0x00);                 // opaque data
            }
            var vxi = list.ToArray();
            list.Clear();
            // RPC
            uint rpcLength = (uint)(40 + vxi.Length);
            rpcLength |= 0x80000000;              // last fragment: yes

            WriteUInt32(list, rpcLength);
            WriteUInt32(list, xid);
            WriteUInt32(list, 0);       // Message Type: Call (0)
            WriteUInt32(list, 2);       // RPC Version: 2
            WriteUInt32(list, 395183);  // Program: VXI-11 Core (395183)
            WriteUInt32(list, 1);       // Program Version: 1
            WriteUInt32(list, 10);      // Procedure: CREATE_LINK (10)
            WriteUInt32(list, 0);       // Credentials flavor: 0
            WriteUInt32(list, 0);       // Credentials length: 0
            WriteUInt32(list, 0);       // Verifier flavor: 0
            WriteUInt32(list, 0);       // Verifier length: 0
            list.AddRange(vxi);
            return list.ToArray();
        }

        private static byte[] CreateRpcLinkDestroy(uint xid, uint linkId)
        {
            var list = new List<byte>();
            // VXI-11
            WriteUInt32(list, linkId);      // Link ID: 0xXXXXXXXX
            var vxi = list.ToArray();
            list.Clear();

            // RPC
            uint rpcLength = (uint)(40 + vxi.Length);
            rpcLength |= 0x80000000;              // last fragment: yes

            WriteUInt32(list, rpcLength);
            WriteUInt32(list, xid);
            WriteUInt32(list, 0);       // Message Type: Call (0)
            WriteUInt32(list, 2);       // RPC Version: 2
            WriteUInt32(list, 395183);  // Program: VXI-11 Core (395183)
            WriteUInt32(list, 1);       // Program Version: 1
            WriteUInt32(list, 23);      // Procedure: DESTROY_LINK (23)
            WriteUInt32(list, 0);       // Credentials flavor: 0
            WriteUInt32(list, 0);       // Credentials length: 0
            WriteUInt32(list, 0);       // Verifier flavor: 0
            WriteUInt32(list, 0);       // Verifier length: 0
            list.AddRange(vxi);
            return list.ToArray();
        }

        private static byte[] CreateRpcClear(uint xid, uint linkId, uint timeOut)
        {
            var list = new List<byte>();
            // VXI-11
            WriteUInt32(list, linkId);      // Link ID: 0xXXXXXXXX
            WriteUInt32(list, 0x00000000);  // Flags: 0x00000000
            WriteUInt32(list, 0);           // I/O Timeout: 0
            WriteUInt32(list, timeOut);     // Lock Timeout: 60000
            var vxi = list.ToArray();
            list.Clear();

            // RPC
            uint rpcLength = (uint)(40 + vxi.Length);
            rpcLength |= 0x80000000;              // last fragment: yes

            WriteUInt32(list, rpcLength);
            WriteUInt32(list, xid);
            WriteUInt32(list, 0);       // Message Type: Call (0)
            WriteUInt32(list, 2);       // RPC Version: 2
            WriteUInt32(list, 395183);  // Program: VXI-11 Core (395183)
            WriteUInt32(list, 1);       // Program Version: 1
            WriteUInt32(list, 15);      // Procedure: DEVICE_CLEAR (15)
            WriteUInt32(list, 0);       // Credentials flavor: 0
            WriteUInt32(list, 0);       // Credentials length: 0
            WriteUInt32(list, 0);       // Verifier flavor: 0
            WriteUInt32(list, 0);       // Verifier length: 0
            list.AddRange(vxi);
            return list.ToArray();
        }

        private static byte[] CreateRpcWrite(uint xid, uint linkId, byte[] buffer, int index, int length, uint timeOut)
        {
            var list = new List<byte>();
            // VXI-11
            WriteUInt32(list, linkId);      // Link ID: 0xXXXXXXXX
            WriteUInt32(list, timeOut);     // I/O Timeout: 60000
            WriteUInt32(list, 0);           // Lock Timeout: 0
            WriteUInt32(list, 0x00000008);  // Flags: 0x00000008 (END) [ ]
            uint vxiLength = (uint)length;
            //vxiLength += 4 - (vxiLength % 4);
            WriteUInt32(list, vxiLength);   // Data length
            list.AddRange(buffer.Skip(index).Take(length)); // Data content
            for (var i = 0; i < 4-(vxiLength%4); i++)
            {
                list.Add(0x00);             // fill bytes: opaque data
            }
            var vxi = list.ToArray();
            list.Clear();

            // RPC
            uint rpcLength = (uint)(40 + vxi.Length);
            rpcLength |= 0x80000000;              // last fragment: yes

            WriteUInt32(list, rpcLength);
            WriteUInt32(list, xid);
            WriteUInt32(list, 0);       // Message Type: Call (0)
            WriteUInt32(list, 2);       // RPC Version: 2
            WriteUInt32(list, 395183);  // Program: VXI-11 Core (395183)
            WriteUInt32(list, 1);       // Program Version: 1
            WriteUInt32(list, 11);      // Procedure: DEVICE_WRITE (11)
            WriteUInt32(list, 0);       // Credentials flavor: 0
            WriteUInt32(list, 0);       // Credentials length: 0
            WriteUInt32(list, 0);       // Verifier flavor: 0
            WriteUInt32(list, 0);       // Verifier length: 0
            list.AddRange(vxi);
            return list.ToArray();
        }

        private static byte[] CreateRpcRead(uint xid, uint linkId, uint size, uint timeOut)
        {
            var list = new List<byte>();
            // VXI-11
            WriteUInt32(list, linkId);      // Link ID: 0xXXXXXXXX
            WriteUInt32(list, size);        // Size: 1024
            WriteUInt32(list, timeOut);     // I/O Timeout:  60000
            WriteUInt32(list, 0);           // Lock Timeout: 0
            WriteUInt32(list, 0x00000000);  // Flags: 0x00000000
            WriteUInt32(list, 0x0000000a);  // Termination Character: 0x0a
            var vxi = list.ToArray();
            list.Clear();
            // RPC
            uint rpcLength = (uint)(40 + vxi.Length);
            rpcLength |= 0x80000000;              // last fragment: yes

            WriteUInt32(list, rpcLength);
            WriteUInt32(list, xid);
            WriteUInt32(list, 0);       // Message Type: Call (0)
            WriteUInt32(list, 2);       // RPC Version: 2
            WriteUInt32(list, 395183);  // Program: VXI-11 Core (395183)
            WriteUInt32(list, 1);       // Program Version: 1
            WriteUInt32(list, 12);      // Procedure: DEVICE_READ (12)
            WriteUInt32(list, 0);       // Credentials flavor: 0
            WriteUInt32(list, 0);       // Credentials length: 0
            WriteUInt32(list, 0);       // Verifier flavor: 0
            WriteUInt32(list, 0);       // Verifier length: 0
            list.AddRange(vxi);
            return list.ToArray();
        }

        private static void WriteUInt32(List<byte> list, uint value)
        {
            list.Add((byte)(value >> 24));
            list.Add((byte)(value >> 16));
            list.Add((byte)(value >> 8));
            list.Add((byte)(value >> 0));
        }

        private static uint ReadUInt32(byte[] buffer, int offset)
        {
            return (uint)(
                (buffer[offset + 0] << 24) |
                (buffer[offset + 1] << 16) |
                (buffer[offset + 2] << 8) |
                (buffer[offset + 3] << 0));
        }
    }
}
